
// fx - A compositing library representation.

// Copyright (C) 2005-2006 Visual Media FX Ltd.
// Released under the LGPL.
// For more information, see http://www.openlibraries.org.

#ifdef HAVE_CONFIG_H
#include <openlibraries_global_config.hpp>
#endif

#include <openpluginlib/pl/openpluginlib.hpp>
#include <openpluginlib/pl/utf8_utils.hpp>

#include <openeffectslib/plugins/openmedialib_source/media_clip_operator.hpp>

namespace fs  = boost::filesystem;
namespace il  = olib::openimagelib::il;
namespace ml  = olib::openmedialib::ml;
namespace opl = olib::openpluginlib;

namespace olib { namespace openeffectslib { namespace plugins {

namespace
{
	media_clip_operator::container field_creator( )
	{
		typedef media_clip_operator::container	container;
		typedef media_clip_operator::value_type	value_type;
		
		container m;

		m.insert( value_type( L"InPlayable",	fx::v_node_ptr( ) ) );
		m.insert( value_type( L"in",			fx::v_frame_position( ) ) );
		m.insert( value_type( L"out",			fx::v_frame_position( ) ) );
		m.insert( value_type( L"OutColor0",		fx::v_node_ptr( ) ) );
		m.insert( value_type( L"OutAudio0",		fx::v_node_ptr( ) ) );

		return m;
	}
	
	media_clip_operator::param_container param_creator( )
	{
		typedef media_clip_operator::param_container	param_container;
		typedef media_clip_operator::param_value_type	param_value_type;

		param_container m;
		
		m.insert( param_value_type( L"InPlayable",		fx::v_node_typename ) );
		m.insert( param_value_type( L"in",				fx::v_frame_position_typename ) );
		m.insert( param_value_type( L"out",				fx::v_frame_position_typename ) );
		m.insert( param_value_type( L"OutColor0",		fx::v_out_color_typename ) );
		m.insert( param_value_type( L"OutAudio0",		fx::v_out_audio_typename ) );
		
		return m;
	}
}

media_clip_operator::media_clip_operator( )
	: n_operator_at( field_creator( ), param_creator( ) )
	, playable_( 0 )
	, in_( 0 )
	, out_( 0 )
{ }

bool media_clip_operator::init( )
{
	fx::v_node_ptr* node = pointer<fx::v_node_ptr>( L"InPlayable" );

	if ( node->value( ) )
	{
		op_ = node->value( );
		playable_ = fx::make_playable( node->value( ) );
		in_ = pointer<fx::v_frame_position>( L"in" );
		out_ = pointer<fx::v_frame_position>( L"out" );
		in_->get( ) = 0;
		if ( playable_ )
			out_->get( ) = playable_->length( );
	}

	return playable_;
}

bool media_clip_operator::set_playable( fx::n_operator_ptr op )
{
	fx::v_node_ptr* node = pointer<fx::v_node_ptr>( L"InPlayable" );
	node->get( ) = op;
	return init( );
}

int media_clip_operator::in( ) const
{
	return playable_ ? in_->value( ) : 0;
}

int media_clip_operator::out( ) const
{
	return playable_ ? out_->value( ) : 0;
}

void media_clip_operator::set_in( int in )
{
	if ( playable_ )
		in_->get( ) = in;
}

void media_clip_operator::set_out( int out )
{
	if ( playable_ )
		out_->get( ) = out;
}

int media_clip_operator::source_length( ) const
{
	return playable_ ? playable_->length( ) : 0;
}

int media_clip_operator::length( ) const
{
	return playable_ ? out_->value( ) - in_->value( ) : 0;
}

double media_clip_operator::fps( ) const
{
	return playable_ ? playable_->fps( ) : 0;
}

int media_clip_operator::position( ) const
{
	return playable_ ? playable_->position( ) - in_->value( ) : 0;
}

int media_clip_operator::position_image( ) const
{
	return playable_ ? playable_->position_image( ) - in_->value( ) : 0;
}

int media_clip_operator::position_audio( ) const
{
	return playable_ ? playable_->position_audio( ) - in_->value( ) : 0;
}

void media_clip_operator::seek( int position, bool relative )
{
	if ( playable_ )
	{
		clamp_points( );

		if ( relative )
			playable_->seek( position, relative );
		else
			playable_->seek( position + in_->value( ) );

		if ( playable_->position( ) != clamp_position( playable_->position( ) ) )
			playable_->seek( clamp_position( playable_->position( ) ) );
	}
}

void media_clip_operator::seek_image( int position, bool relative )
{
	if ( playable_ )
	{
		clamp_points( );

		if ( relative )
			playable_->seek_image( position, relative );
		else
			playable_->seek_image( position + in_->value( ) );

		if ( playable_->position_image( ) != clamp_position( playable_->position_image( ) ) )
			playable_->seek_image( clamp_position( playable_->position_image( ) ) );
	}
}

void media_clip_operator::seek_audio( int position, bool relative )
{
	if ( playable_ )
	{
		clamp_points( );

		if ( relative )
			playable_->seek_audio( position, relative );
		else
			playable_->seek_audio( position + in_->value( ) );

		if ( playable_->position_image( ) != clamp_position( playable_->position_audio( ) ) )
			playable_->seek_audio( clamp_position( playable_->position_audio( ) ) );
	}
}

fx::n_operator_ptr media_clip_operator::op_image( ) const
{
	return op_;
}

fx::n_operator_ptr media_clip_operator::op_audio( ) const
{
	return op_;
}

void media_clip_operator::clamp_points( )
{
	if ( out_->value( ) < in_->value( ) )
	{
		int t = in_->value( );
		in_->get( ) = out_->value( );
		out_->get( ) = t;
	}

	if ( in_->value( ) < 0 )
		in_->get( ) = 0;
	else if ( in_->value( ) >= playable_->length( ) )
		in_->get( ) = playable_->length( ) - 1;

	if ( out_->value( ) < 0 )
		out_->get( ) = 1;
	else if ( out_->value( ) > playable_->length( ) )
		out_->get( ) = playable_->length( );
}

int media_clip_operator::clamp_position( int position )
{
	if ( position < in_->value( ) )
		position = in_->value( );
	else if ( position >= out_->value( ) )
		position = out_->value( ) - 1;
	return position;
}

} } }
