
// fx - A compositing library representation.

// Copyright (C) 2005-2006 Visual Media FX Ltd.
// Released under the LGPL.
// For more information, see http://www.openlibraries.org.

#ifdef HAVE_CONFIG_H
#include <openlibraries_global_config.hpp>
#endif

#include <openpluginlib/pl/openpluginlib.hpp>
#include <openpluginlib/pl/utf8_utils.hpp>

#include <openeffectslib/plugins/openmedialib_source/media_source_operator.hpp>

namespace fs  = boost::filesystem;
namespace il  = olib::openimagelib::il;
namespace ml  = olib::openmedialib::ml;
namespace opl = olib::openpluginlib;

namespace olib { namespace openeffectslib { namespace plugins {

namespace
{
	media_source_operator::container field_creator( )
	{
		typedef media_source_operator::container	container;
		typedef media_source_operator::value_type	value_type;
		
		container m;

		m.insert( value_type( L"url",			opl::v_url( ) ) );
		m.insert( value_type( L"imagePosition",	fx::v_frame_position( ) ) );
		m.insert( value_type( L"audioPosition",	fx::v_frame_position( ) ) );
		m.insert( value_type( L"frameCount",	fx::v_frame_position( ) ) );
		m.insert( value_type( L"frameRate",		opl::v_double( ) ) );
		m.insert( value_type( L"OutColor0",		fx::v_node_ptr( ) ) );
		m.insert( value_type( L"OutAudio0",		fx::v_node_ptr( ) ) );

		return m;
	}
	
	media_source_operator::param_container param_creator( )
	{
		typedef media_source_operator::param_container	param_container;
		typedef media_source_operator::param_value_type	param_value_type;

		param_container m;
		
		m.insert( param_value_type( L"url",				opl::v_url_typename ) );
		m.insert( param_value_type( L"imagePosition",	fx::v_frame_position_typename ) );
		m.insert( param_value_type( L"audioPosition",	fx::v_frame_position_typename ) );
		m.insert( param_value_type( L"frameCount",		fx::v_frame_position_typename ) );
		m.insert( param_value_type( L"frameRate",		opl::v_double_typename ) );
		m.insert( param_value_type( L"OutColor0",		fx::v_out_color_typename ) );
		m.insert( param_value_type( L"OutAudio0",		fx::v_out_audio_typename ) );
		
		return m;
	}
}

namespace
{
	ml::input_type_ptr create_input( const std::string& file )
	{ return ml::create_input( file.c_str( ) ); }
}

media_source_operator::media_source_operator( )
	: n_operator_gpu( field_creator( ), param_creator( ) )
	, id_( 0 )
	, target_( GL_TEXTURE_2D )
	, old_phy_w_( 0 )
	, old_phy_h_( 0 )
	, frame_count_( 0 )
	, frame_rate_( 0 )
	, image_position_( 0 )
	, audio_position_( 0 )
{ }

bool media_source_operator::init( )
{
	opl::v_url* url = pointer<opl::v_url>( L"url" );
	input_ = ml::create_input( url->value( ).string( ).c_str( ) );
	if ( input_ )
	{
		frame_count_ = pointer<fx::v_frame_position>( L"frameCount" );
		frame_rate_ = pointer<opl::v_double>( L"frameRate" );
		image_position_ = pointer<fx::v_frame_position>( L"imagePosition" );
		audio_position_ = pointer<fx::v_frame_position>( L"audioPosition" );

		frame_count_->get( ) = input_->get_frames( );
		frame_rate_->get( ) = input_->fps( );
		image_position_->get( ) = 0;
		audio_position_->get( ) = 0;
	}
	return input_;
}

bool media_source_operator::update_image( )
{
	if ( input_ == 0 && !init( ) ) return false;

	if( !fetch_frame( input_, image_position_->value( ) ) || !rb( ).initialised( ) ) return false;
	image_position_->get( ) = input_->get_position( );

	glDisable( GL_DEPTH_TEST );
	glDisable( GL_LIGHTING );
	glDisable( GL_CULL_FACE );

	rb( ).begin_render_to( );

	int rb_w		= width( );
	int rb_h		= height( );
	float p_max_s	= rb( ).physical_max_s( );
	float p_max_t	= rb( ).physical_max_t( );

	glEnable( target_ );
	glBindTexture( target_, id_ );
		
	glBegin( GL_QUADS );
		glTexCoord2f( 0.0f, 0.0f );			glVertex2i( 0, 0 );
		glTexCoord2f( p_max_s, 0.0f );		glVertex2i( rb_w, 0 );
		glTexCoord2f( p_max_s, p_max_t );	glVertex2i( rb_w, rb_h );
		glTexCoord2f( 0.0f, p_max_t );		glVertex2i( 0, rb_h );
	glEnd( );

	glBindTexture( target_, 0 );
	glDisable( target_ );

	rb( ).end_render_to( );

	return true;
}

bool media_source_operator::update_audio( )
{
	if ( input_ == 0 && !init( ) ) return false;

	input_->seek( audio_position_->value( ) );
	ml::frame_type_ptr frame = input_->fetch( );
	if( !frame ) return false;
	set_ab( frame->get_audio( ) );

	audio_position_->get( ) = input_->get_position( );

	return true;
}

int media_source_operator::length( ) const
{
	return frame_count_->value( );
}

double media_source_operator::fps( ) const
{
	return frame_rate_->value( );
}

int media_source_operator::position( ) const
{
	return image_position_->value( );
}

int media_source_operator::position_image( ) const
{
	return image_position_->value( );
}

int media_source_operator::position_audio( ) const
{
	return audio_position_->value( );
}

void media_source_operator::seek( int position, bool relative )
{
	seek_image( position, relative );
	seek_audio( position, relative );
}

void media_source_operator::seek_image( int position, bool relative )
{
	int next = relative ? image_position_->value( ) + position : position;
	if ( next < 0 ) next = 0;
	if ( next >= length( ) ) next = length( ) - 1;
	image_position_->get( ) = next;
}

void media_source_operator::seek_audio( int position, bool relative )
{
	int next = relative ? audio_position_->value( ) + position : position;
	if ( next < 0 ) next = 0;
	if ( next >= length( ) ) next = length( ) - 1;
	audio_position_->get( ) = next;
}

bool media_source_operator::fetch_frame( ml::input_type_ptr input, frame_position_t position )
{
	if( !input ) return false;
	
	input->seek( position );

	ml::frame_type_ptr frame = input->fetch( );
	if( !frame || !frame->get_image( ) ) return false;
	
	il::image_type_ptr im = frame->get_image( );
	
	im = il::conform( im, il::cropped );
	im = il::deinterlace( im );
	im = il::convert( im, L"r8g8b8" );

	int phy_w = im->width( );
	int phy_h = im->height( );

	float dummy_tex_w, dummy_tex_h;
	if( !opl::texture_target( phy_w, phy_h, target_, dummy_tex_w, dummy_tex_h ) ) return false;

	typedef std::pair<GLenum, GLenum> internal_and_base_format;
	internal_and_base_format pf = opl::pf_to_gl_format( im->pf( ) );

	if( id_ == 0 )
		glGenTextures( 1, &id_ );

	glPixelStorei( GL_UNPACK_ALIGNMENT, 4 );
	glBindTexture( target_, id_ );	
	glTexParameteri( target_, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
	glTexParameteri( target_, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );

	if ( phy_w != old_phy_w_ || phy_h != old_phy_h_ )
	{
		glTexParameteri( target_, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
		glTexParameteri( target_, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
		glTexImage2D( target_, 0, pf.first, phy_w, phy_h, 0, pf.second, GL_UNSIGNED_BYTE, im->data( ) );
		old_phy_w_ = phy_w;
		old_phy_h_ = phy_h;
		
		rb( ).init( "", int( float( phy_h ) * frame->aspect_ratio( ) ), phy_h, phy_w, phy_h ); // TODO: can optimise further.
	}
	else
	{
		glTexSubImage2D( target_, 0, 0, 0, phy_w, phy_h, pf.second, GL_UNSIGNED_BYTE, im->data( ) );
	}
	
	glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
	
	return true;
}

} } }
