
// HDR - A HDR plugin to il.

// Copyright (C) 2005 Visual Media FX Ltd.
// Released under the LGPL.
// For more information, see http://www.openlibraries.org.

#ifdef WIN32
#include <windows.h>
#include <gdiplus.h>

#ifdef min
#undef min
#endif
#endif // WIN32

#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <limits>
#include <vector>
#include <string>

#include <boost/filesystem/fstream.hpp>

#include <openimagelib/plugins/hdr/hdr_plugin.hpp>

namespace il = olib::openimagelib::il;
namespace fs = boost::filesystem;

namespace olib { namespace openimagelib { namespace plugins { namespace HDR {

typedef il::image<unsigned char, il::surface_format> image_type;

namespace
{
	void destroy( image_type* im )
	{ delete im; }

	bool Read_s( fs::ifstream& file, char* s, std::streamsize size, std::streamsize max )
	{
#if _MSC_VER >= 1400
		file._Read_s( s, size, max );
#else
		file.read( s, size );
#endif
			
		return !file.fail( );
	}

	void float2rgbe( float red, float green, float blue, unsigned char rgbe[ 4 ] )
	{
		float v = red;
		if( v < green ) v = green;
		if( v < blue  ) v = blue;
		
		if( v < std::numeric_limits<float>::min( ) )
		{
			rgbe[ 0 ] = rgbe[ 1 ] = rgbe[ 2 ] = rgbe[ 3 ] = 0;
		}
		else
		{
			int exponent;
			v = frexp( v, &exponent );
			
			rgbe[ 0 ] = static_cast<unsigned char>( red   * v );
			rgbe[ 1 ] = static_cast<unsigned char>( green * v );
			rgbe[ 2 ] = static_cast<unsigned char>( blue  * v );
			rgbe[ 3 ] = exponent + 128;
		}	
	}
	
	void rgbe2float( unsigned char rgbe[ 4 ], float& red, float& green, float& blue )
	{
		if( rgbe[ 3 ] )
		{
			float f = ldexp( 1.0f, rgbe[ 3 ] - ( 128 + 8 ) );
			
			red	  = rgbe[ 0 ] * f;
			green = rgbe[ 1 ] * f;
			blue  = rgbe[ 2 ] * f;
		}
		else
		{
			red = green = blue = 0.0f;
		}
	}
	
	bool read_rgbe_pixels_raw( fs::ifstream& file, float* data, int width, int height )
	{
		char rgbe[ 4 ];
		
		int numpixels = width * height;
		while( numpixels-- )
		{
			if( !Read_s( file, rgbe, sizeof( rgbe ), sizeof( rgbe ) ) )
				return false;
				
			rgbe2float( ( unsigned char* ) rgbe, data[ 0 ], data[ 1 ], data[ 2 ] );
			
			data += 3;
		}
		
		return true;
	}
	
	bool read_rgbe_pixels( fs::ifstream& file, float* data, int width, int height )
	{
		if( width < 8 || width > 0x7FFF )
			return read_rgbe_pixels_raw( file, data, width, height );
		
		std::vector<unsigned char> line;
		line.resize( width * 4 );
		
		unsigned char rgbe[ 4 ];
		for( int i = 0; i < height; ++i )
		{
			if( !Read_s( file, ( char* ) rgbe, sizeof( rgbe ), sizeof( rgbe ) ) )
				return false;
				
			if( ( rgbe[ 0 ] != 2 ) || ( rgbe[ 1 ] != 2 ) || ( rgbe[ 2 ] & 0x80 ) )
			{
				rgbe2float( rgbe, data[ 0 ], data[ 1 ], data[ 2 ] );
				
				data += 3;
				
				return read_rgbe_pixels_raw( file, data, width, height );
			}
			
			if( ( ( ( int ) rgbe[ 2 ] ) << 8 | rgbe[ 3 ] ) != width )
				return false;
				
			for( int j = 0; j < 4; ++j )
			{
				typedef std::vector<unsigned char>::const_iterator	const_iterator;
				typedef std::vector<unsigned char>::iterator		iterator;
				
				iterator		I = line.begin( );
				const_iterator	J = line.begin( ) + ( j + 1 ) * width;
				while( I != J )
				{
					unsigned char buf[ 2 ];
					if( !Read_s( file, ( char* ) buf, sizeof( buf ), sizeof( buf ) ) )
						return false;
						
					if( buf[ 0 ] > 128 )
					{
						int count = buf[ 0 ] - 128;
						if( !count || ( count > ( J - I ) ) ) return false;
						
						while( count-- ) *I++ = buf[ 1 ];
					}
					else
					{
						int count = buf[ 0 ];
						if( !count || ( count > ( J - I ) ) ) return false;
						
						*I++ = buf[ 1 ];
						if( --count > 0 )
						{
							if( !Read_s( file, ( char* ) &line[ 0 ], sizeof( count ), sizeof( count ) ) )
								return false;
						}
						
						I += count;
					}
				}
				
				for( int k = 0; k < width; ++k )
				{
					rgbe[ 0 ] = line[ i + 0 * width ];
					rgbe[ 1 ] = line[ i + 1 * width ];
					rgbe[ 2 ] = line[ i + 2 * width ];
					rgbe[ 3 ] = line[ i + 3 * width ];
					
					rgbe2float( rgbe, data[ 0 ], data[ 1 ], data[ 2 ] );
					
					data += 3;
				}
			}
		}
		
		return true;
	}

	boost::shared_ptr<image_type> rgbe_to_image_type( int width, int height )
	{
		using namespace olib::openimagelib::il;

		typedef boost::shared_ptr<image_type>		image_type_ptr;
		typedef image<unsigned char, r32g32b32f>	r32g32b32f_image_type;

		return image_type_ptr( new image_type( r32g32b32f_image_type( width, height, 1 ) ), destroy );
	}

	HDR_plugin::image_type_ptr load_hdr( const boost::filesystem::path& path )
	{
		using namespace olib::openimagelib::il;
		
		typedef boost::shared_ptr<image_type> image_type_ptr;

		int width, height;

#define RGBE_VALID_PROGRAMTYPE 0x01
#define RGBE_VALID_GAMMA       0x02
#define RGBE_VALID_EXPOSURE    0x04

		struct RgbeInfo
		{
			int   valid;
			char  program_type[ 16 ];
			float gamma;
			float exposure;
			char  format[ 16 ];
		};

		fs::ifstream file( path, std::ios::in | std::ios::binary );
		if( !file.is_open( ) )
			return HDR_plugin::image_type_ptr( static_cast<image_type*>( 0 ) );

		RgbeInfo rgbe_info = { 0, { 0 }, 1.0f, 1.0f, { 0 } };

		char buffer[ 128 ];
		if( !Read_s( file, buffer, sizeof( buffer ), sizeof( buffer ) ) )
			return HDR_plugin::image_type_ptr( static_cast<image_type*>( 0 ) );

		if( buffer[ 0 ] != '#' || buffer[ 1 ] != '?' )
			return HDR_plugin::image_type_ptr( static_cast<image_type*>( 0 ) );

		rgbe_info.valid |= RGBE_VALID_PROGRAMTYPE;
		for( size_t i = 0; i < sizeof( rgbe_info.program_type ) - 1; ++i )
		{
			if( buffer[ i + 2 ] == 0 || isspace( buffer[ i + 2 ] ) )
				break;

			rgbe_info.program_type[ i ] = buffer[ i + 2 ];
		}
		rgbe_info.program_type[ 15 ] = 0;

		if( !Read_s( file, buffer, sizeof( buffer ), sizeof( buffer ) ) )
			return HDR_plugin::image_type_ptr( static_cast<image_type*>( 0 ) );

		if( !buffer[ 0 ] )
			return HDR_plugin::image_type_ptr( static_cast<image_type*>( 0 ) );

		for( ;; )
		{
			if( !strcmp( buffer, "\n" ) ) break;
			else if( !strcmp( buffer, "FORMAT=32-bit_rle_rgbe\n" ) )
			{
#if _MSC_VER >= 1400
				strncpy_s( rgbe_info.format, "32-bit_rle_rgbe", strlen( "32-bit_rle_rgbe" ) );
#else
				strncpy( rgbe_info.format, "32-bit_rle_rgbe", strlen( "32-bit_rle_rgbe" ) );
#endif
			}
#if _MSC_VER >= 1400
			else if( sscanf_s( buffer, "GAMMA=%g", &rgbe_info.gamma ) == 1 )
#else
			else if( sscanf( buffer, "GAMMA=%g", &rgbe_info.gamma ) == 1 )
#endif
			{
				rgbe_info.valid |= RGBE_VALID_GAMMA;
			}
#if _MSC_VER >= 1400
			else if( sscanf_s( buffer, "EXPOSURE=%g", &rgbe_info.exposure ) == 1 )
#else
			else if( sscanf( buffer, "EXPOSURE=%g", &rgbe_info.exposure ) == 1 )
#endif
			{
				rgbe_info.valid |= RGBE_VALID_EXPOSURE;
			}

			if( !Read_s( file, buffer, sizeof( buffer ), sizeof( buffer ) ) )
				return HDR_plugin::image_type_ptr( static_cast<image_type*>( 0 ) );
		}

#undef RGBE_VALID_PROGRAMTYPE
#undef RGBE_VALID_GAMMA
#undef RGBE_VALID_EXPOSURE

		if( !Read_s( file, buffer, sizeof( buffer ), sizeof( buffer ) ) )
			return HDR_plugin::image_type_ptr( static_cast<image_type*>( 0 ) );

#if _MSC_VER >= 1400
		if( sscanf_s( buffer, "-Y %d +X %d", &width, &height ) < 2 )
#else
		if( sscanf( buffer, "-Y %d +X %d", &width, &height ) < 2 )
#endif
			return HDR_plugin::image_type_ptr( static_cast<image_type*>( 0 ) );
			
		HDR_plugin::image_type_ptr image = rgbe_to_image_type( width, height );
		if( !image )
			return HDR_plugin::image_type_ptr( static_cast<image_type*>( 0 ) );
			
		if( !read_rgbe_pixels( file, reinterpret_cast<float*>( image->data( ) ), width, height ) )
			return HDR_plugin::image_type_ptr( static_cast<image_type*>( 0 ) );
			
		return HDR_plugin::image_type_ptr( image );
	}
}
			
HDR_plugin::image_type_ptr HDR_plugin::load( const boost::filesystem::path& path )
{ return HDR_plugin::image_type_ptr( load_hdr( path ) ); }

bool HDR_plugin::store( const boost::filesystem::path& path, const HDR_plugin::image_type_ptr& )
{ return false; }

} } } }
