
// TSTO - A TSTO plugin to sg.

// Copyright (C) 2005-2006 Visual Media FX Ltd.
// Released under the LGPL.
// For more information, see http://www.openlibraries.org.

#include <algorithm>
#include <string>
#include <vector>

#ifndef NDEBUG
//#define BOOST_SPIRIT_DEBUG
#endif

#include <boost/filesystem/fstream.hpp>
#include <boost/tokenizer.hpp>

#include <boost/lexical_cast.hpp>

#include <boost/spirit/core.hpp>
#include <boost/spirit/utility.hpp>
#include <boost/spirit/symbols.hpp>

#include <openpluginlib/pl/string.hpp>

#include <openobjectlib/plugins/tsto/tsto_plugin.hpp>

#include <openobjectlib/sg/meshes.hpp>
#include <openobjectlib/sg/spatial.hpp>

namespace fs		= boost::filesystem;
namespace spirit	= boost::spirit;
namespace opl		= olib::openpluginlib;
namespace sg		= olib::openobjectlib::sg;

namespace olib { namespace openobjectlib { namespace plugins { namespace TSTO {

namespace
{
	struct grammar_state
	{
		typedef std::vector<float>	vec_floats;
		typedef std::vector<int>	vec_ints;
		
		vec_floats v, n, t, c;
		vec_ints tris;
	};

	struct extractor
	{
		template<class T>
		void extract( const opl::string& str, std::vector<T>& v )
		{ v.push_back( boost::lexical_cast<T>( str.c_str( ) ) ); }
	};

	struct geometry_action
	{
		explicit geometry_action( grammar_state::vec_floats& v )
			: v_( v )
		{ }
		
		void operator( )( const char* first, const char* last ) const
		{ extractor( ).extract( opl::string( first, last ), v_ ); }
		
		grammar_state::vec_floats& v_;
	};

	struct topology_action
	{
		explicit topology_action( grammar_state::vec_ints& v )
			: v_( v )
		{ }

		void operator( )( const char* first, const char* last ) const
		{ extractor( ).extract( opl::string( first, last ), v_ ); }
		
		grammar_state::vec_ints& v_;
	};

	///////////////////////////////////////////////////////////////
	// skip grammar (whitespace and comments)
	struct skip_grammar : public spirit::grammar<skip_grammar>
	{
		template<class ScannerT>
		struct definition
		{
			definition( const skip_grammar& /*self*/)
				: COMMA( ',' )
			{
				skip
					=	spirit::space_p
					|	spirit::comment_p( "//" )
					|	spirit::comment_p( "/*", "*/" )
					|	COMMA
					;
					
#		ifndef NDEBUG
				BOOST_SPIRIT_DEBUG_NODE( skip );
#		endif
			}

			spirit::chlit<> COMMA;
			
			spirit::rule<ScannerT> skip;
		
			const spirit::rule<ScannerT>& start( ) const
			{ return skip; }
		};
	};

	///////////////////////////////////////////////////////////////
	//
	struct TSTO_grammar : public spirit::grammar<TSTO_grammar>
	{
		static const int RESERVE_SIZ = 1000;

		TSTO_grammar( grammar_state& eval )
			: eval_( eval )
		{
			eval_.v.reserve( RESERVE_SIZ );
			eval_.n.reserve( RESERVE_SIZ );
			eval_.t.reserve( RESERVE_SIZ );
			eval_.c.reserve( RESERVE_SIZ );
			eval_.tris.reserve( RESERVE_SIZ );
		}
		
		template<class ScannerT>
		struct definition
		{
			//////////////////////////////////////////////////
			// grammar definition
			definition( const TSTO_grammar& self )
				: LPAREN( '(' )
				, RPAREN( ')' )
				, LBRACE( '{' )
				, RBRACE( '}' )
				, SEMICOLON( ';' )
				, DOUBLEQUOTE( '"' )
			{
				using namespace boost::spirit;
				
				VERTEX3F	= strlit<>( "Vertex3f" );
				COLOR3F		= strlit<>( "Color3f" );
				SPECULAR3F	= strlit<>( "Specular3f" );
				NORMAL3F	= strlit<>( "Normal3f" );
				TEXCOORD3F	= strlit<>( "TexCoord3f" );
				TRIANGLE	= strlit<>( "Triangle" );
				USEMATERIAL	= strlit<>( "UseMaterial" );
				OBJECTFIT	= strlit<>( "ObjectFit" );
				OBJECT		= strlit<>( "Object" );

				IDENTIFIER
					=	lexeme_d
						[ 
							( ( alpha_p | '_' | '$' ) >> *( alnum_p | '_' | '$' | '-' | '.' ) ) - ( keywords >> anychar_p - ( alnum_p | '_' | '$' ) )
						]
					;

				STRING
					=	DOUBLEQUOTE >> IDENTIFIER >> DOUBLEQUOTE
					;
					
				INTEGER_DECIMAL_CONSTANT
					=	lexeme_d[ +digit_p ]
					;
					
				INTEGER_CONSTANT
					=	INTEGER_DECIMAL_CONSTANT
					;

				FLOAT_CONSTANT_1    // 12345[eE][+-]123[lLfF]?
					=	lexeme_d
						[	
							+digit_p
						>>	( chlit<>( 'e' )  | chlit<>( 'E' ) )
						>>	!sign_p
						>>	+digit_p
						>>	!as_lower_d[ chlit<>( 'l' ) | chlit<>( 'f' ) ]
						]
					;

				FLOAT_CONSTANT_2    // .123([[eE][+-]123)?[lLfF]?
					=	lexeme_d
						[ 
							
							*digit_p
						>>	chlit<>('.')
						>>	+digit_p
						>>	!(  ( chlit<>( 'e' ) | chlit<>( 'E' ) )
							 >> !sign_p
							 >> +digit_p
							 )
						>> !as_lower_d[ chlit<>( 'l' ) | chlit<>( 'f' ) ]
						]
					;

				FLOAT_CONSTANT_3    // 12345.([[eE][+-]123)?[lLfF]?
					=	lexeme_d
						[ 
								+digit_p
							>>	chlit<>( '.' )
							>>	*digit_p
							>>	!(  ( chlit<>( 'e' ) | chlit<>( 'E' ) )
								 >> !sign_p
								 >> +digit_p
								 )
							>>	!as_lower_d[ chlit<>( 'l' ) | chlit<>( 'f' ) ]
						]
					;

				FLOAT_CONSTANT
					=	FLOAT_CONSTANT_1
					|	FLOAT_CONSTANT_2
					|	FLOAT_CONSTANT_3
					;
					
				CONSTANT
					=	!sign_p >> longest_d[ FLOAT_CONSTANT | INTEGER_CONSTANT ]
					;

				vertex3f
					=	VERTEX3F
					>>	LPAREN
					>>		CONSTANT[ geometry_action( ( const_cast<TSTO_grammar&>( self ) ).eval_.v ) ]
					>>		CONSTANT[ geometry_action( ( const_cast<TSTO_grammar&>( self ) ).eval_.v ) ]
					>>		CONSTANT[ geometry_action( ( const_cast<TSTO_grammar&>( self ) ).eval_.v ) ]
					>>		*CONSTANT
					>>	RPAREN
					;

				normal3f
					=	NORMAL3F
					>>	LPAREN
					>>		CONSTANT[ geometry_action( ( const_cast<TSTO_grammar&>( self ) ).eval_.n ) ]
					>>		CONSTANT[ geometry_action( ( const_cast<TSTO_grammar&>( self ) ).eval_.n ) ]
					>>		CONSTANT[ geometry_action( ( const_cast<TSTO_grammar&>( self ) ).eval_.n ) ]
					>>		*CONSTANT
					>>	RPAREN
					;

				color3f
					=	COLOR3F
					>>	LPAREN
					>>		CONSTANT[ geometry_action( ( const_cast<TSTO_grammar&>( self ) ).eval_.c ) ]
					>>		CONSTANT[ geometry_action( ( const_cast<TSTO_grammar&>( self ) ).eval_.c ) ]
					>>		CONSTANT[ geometry_action( ( const_cast<TSTO_grammar&>( self ) ).eval_.c ) ]
					>>		*CONSTANT
					>>	RPAREN
					;

				texcoord3f
					=	TEXCOORD3F
					>>	LPAREN
					>>		CONSTANT[ geometry_action( ( const_cast<TSTO_grammar&>( self ) ).eval_.t ) ]
					>>		CONSTANT[ geometry_action( ( const_cast<TSTO_grammar&>( self ) ).eval_.t ) ]
					>>		CONSTANT[ geometry_action( ( const_cast<TSTO_grammar&>( self ) ).eval_.t ) ]
					>>		*CONSTANT
					>>	RPAREN
					;

				specular3f
					=	SPECULAR3F
					>>	LPAREN
					>>		CONSTANT
					>>		CONSTANT
					>>		CONSTANT
					>>		*CONSTANT
					>>	RPAREN
					;

				objectfit
					=	OBJECTFIT
					>>	LPAREN
					>>		CONSTANT
					>>		CONSTANT
					>>		*CONSTANT
					>>	RPAREN
					;

				triangle
					=	TRIANGLE
					>>	LPAREN
					>>		INTEGER_CONSTANT[ topology_action( ( const_cast<TSTO_grammar&>( self ) ).eval_.tris ) ]
					>>		INTEGER_CONSTANT[ topology_action( ( const_cast<TSTO_grammar&>( self ) ).eval_.tris ) ]
					>>		INTEGER_CONSTANT[ topology_action( ( const_cast<TSTO_grammar&>( self ) ).eval_.tris ) ]
					>>	RPAREN
					;
					
				use_material
					=	USEMATERIAL >> LPAREN >> STRING >> RPAREN
					;

				object_element
					=	(	vertex3f
						|	normal3f
						|	color3f
						|	texcoord3f
						|	specular3f
						|	objectfit
						|	triangle
						|	use_material
						)
					 >> SEMICOLON
					;

				object 
					=	OBJECT >> LPAREN >> STRING >> RPAREN 
					>>	LBRACE
					>>		*object_element
					>>	RBRACE
					;

#		ifndef NDEBUG
				BOOST_SPIRIT_DEBUG_NODE( INTEGER_DECIMAL_CONSTANT );
				BOOST_SPIRIT_DEBUG_NODE( INTEGER_CONSTANT );
				BOOST_SPIRIT_DEBUG_NODE( FLOAT_CONSTANT );
				BOOST_SPIRIT_DEBUG_NODE( FLOAT_CONSTANT_1 );
				BOOST_SPIRIT_DEBUG_NODE( FLOAT_CONSTANT_2 );
				BOOST_SPIRIT_DEBUG_NODE( FLOAT_CONSTANT_3 );
				BOOST_SPIRIT_DEBUG_NODE( CONSTANT );
				BOOST_SPIRIT_DEBUG_NODE( IDENTIFIER );
				BOOST_SPIRIT_DEBUG_NODE( STRING );

				BOOST_SPIRIT_DEBUG_NODE( VERTEX3F );
				BOOST_SPIRIT_DEBUG_NODE( COLOR3F );
				BOOST_SPIRIT_DEBUG_NODE( NORMAL3F );
				BOOST_SPIRIT_DEBUG_NODE( TEXCOORD3F );
				BOOST_SPIRIT_DEBUG_NODE( SPECULAR3F );
				BOOST_SPIRIT_DEBUG_NODE( TRIANGLE );
				BOOST_SPIRIT_DEBUG_NODE( USEMATERIAL );
				BOOST_SPIRIT_DEBUG_NODE( OBJECTFIT );
				BOOST_SPIRIT_DEBUG_NODE( OBJECT );

				BOOST_SPIRIT_DEBUG_NODE( vertex3f );
				BOOST_SPIRIT_DEBUG_NODE( color3f );
				BOOST_SPIRIT_DEBUG_NODE( normal3f );
				BOOST_SPIRIT_DEBUG_NODE( texcoord3f );
				BOOST_SPIRIT_DEBUG_NODE( specular3f );
				BOOST_SPIRIT_DEBUG_NODE( triangle );
				BOOST_SPIRIT_DEBUG_NODE( usematerial );
				BOOST_SPIRIT_DEBUG_NODE( objectfit );
				BOOST_SPIRIT_DEBUG_NODE( object );
				BOOST_SPIRIT_DEBUG_NODE( object_element );
				BOOST_SPIRIT_DEBUG_NODE( use_material );
#		endif
			}

			spirit::symbols<> keywords;
			
			spirit::chlit<> LPAREN, RPAREN, LBRACE, RBRACE, SEMICOLON, DOUBLEQUOTE;
			
			spirit::rule<ScannerT> INTEGER_DECIMAL_CONSTANT, INTEGER_CONSTANT, FLOAT_CONSTANT;
			spirit::rule<ScannerT> FLOAT_CONSTANT_1, FLOAT_CONSTANT_2, FLOAT_CONSTANT_3, CONSTANT;
			spirit::rule<ScannerT> VERTEX3F, COLOR3F, NORMAL3F, TEXCOORD3F, SPECULAR3F;
			spirit::rule<ScannerT> TRIANGLE, USEMATERIAL, OBJECTFIT, OBJECT;
			spirit::rule<ScannerT> IDENTIFIER, STRING;

			spirit::rule<ScannerT> vertex3f, color3f, normal3f, texcoord3f, specular3f;
			spirit::rule<ScannerT> triangle, usematerial, objectfit, object, object_element;
			spirit::rule<ScannerT> use_material;
			
			const spirit::rule<ScannerT>& start( ) const
			{ return object; }
		};
		
		grammar_state& eval_;
	};

	template<typename C, typename V>
	void copy_( const C& c, V& v )
	{
#if _MSC_VER >= 1400
		std::copy( c.begin( ), c.end( ), stdext::checked_array_iterator<typename V::iterator>( v.begin( ), v.size( ) ) );
#else
		std::copy( c.begin( ), c.end( ), v.begin( ) );
#endif
	}

	void add_attribute( const sg::node_ptr& m, const std::vector<float>& v, const opl::wstring& name )
	{
		sg::attribute_array_ptr a( new sg::attribute_array( ) );
		
		opl::m_float* val = a->pointer<opl::m_float>( L"value" );
		opl::v_int32* nc  = a->pointer<opl::v_int32>( L"components" );
		
		val->resize( v.size( ) );
		nc->get( ) = 3;
		
		copy_( v, *val );
	
		sg::v_attribute_array_ptr* attr_ = m->pointer<sg::v_attribute_array_ptr>( name );
		attr_->get( ) = a;
	}

	sg::spatial_ptr build_tsto_graph( const TSTO_grammar& g )
	{
		sg::spatial_ptr s( new sg::shape( ) );
		
		sg::mesh_ptr m( new sg::mesh( ) );
		m->prim_type( ) = sg::mesh::TRIANGLES;
		
		if( !g.eval_.v.empty( ) )
			add_attribute( m, g.eval_.v, L"coord" );

		if( !g.eval_.n.empty( ) )
			add_attribute( m, g.eval_.n, L"normal" );

		if( !g.eval_.c.empty( ) )
			add_attribute( m, g.eval_.c, L"color" );

		if( !g.eval_.t.empty( ) )
		{
			sg::mv_uv_set_ptr* uv_ = m->pointer<sg::mv_uv_set_ptr>( L"uv" );
			
			sg::uv_set_ptr uv_set( new sg::uv_set( ) );
			uv_->push_back( sg::v_uv_set_ptr( uv_set ) );

			add_attribute( uv_set, g.eval_.t, L"uv" );
		}

		if( !g.eval_.tris.empty( ) )
		{
			opl::m_int32* i = m->pointer<opl::m_int32>( L"index" );
			
			i->resize( g.eval_.tris.size( ) );
			copy_( g.eval_.tris, *i );
		}
		
		sg::v_mesh_ptr* vm = s->pointer<sg::v_mesh_ptr>( L"geometry" );
		vm->get( ) = m;
		
		return s;
	}

	bool parse_and_build_tsto_graph( const std::string& tsto, sg::spatial_ptr& root )
	{
		grammar_state eval;
		
		TSTO_grammar tsto_g( eval );
		skip_grammar skip_g;
		
#ifndef NDEBUG
		BOOST_SPIRIT_DEBUG_NODE( tsto_g );
		BOOST_SPIRIT_DEBUG_NODE( skip_g );
#endif

		spirit::parse_info<> result = spirit::parse( tsto.c_str( ), tsto_g, skip_g );
		
		return result.full && ( root = build_tsto_graph( tsto_g ) );
	}

	sg::scene_ptr load_tsto( const fs::path& path )
	{
		sg::scene_ptr sc( new sg::scene( ) );
		
		fs::ifstream file( path, std::ios::in );
		if( !file.is_open( ) )
			return sc;

		file.unsetf( std::ios::skipws );

		std::string tsto( ( std::istreambuf_iterator<char>( file ) ), std::istreambuf_iterator<char>( ) );
		
		sg::spatial_ptr root;
		if( !parse_and_build_tsto_graph( tsto, root ) )
			return sc;
		
		sc->set_root( root );

		return sc;
	}
}

sg::scene_ptr TSTO_plugin::load( const fs::path& path, const opl::string& /*options*/ )
{ return load_tsto( path ); }

bool TSTO_plugin::store( const fs::path& path, const sg::scene_ptr& root )
{ return false; }

} } } }
